/*
  Ncr4x20Vfd Library - Declaration file
  
  Originally created 6 Feb 2011

  This file is in the public domain.  
*/

#ifndef Ncr4X20Vfd_h
#define Ncr4X20Vfd_h

#include <inttypes.h>
#include "Print.h"

class Ncr4X20Vfd : public Print {
public:

  // Special control characters supported by the display.
  static const uint8_t CURSOR_ON_CHARACTER = 0x15;
  static const uint8_t CURSOR_OFF_CHARACTER = 0x16;
  static const uint8_t ALTERNATE_CURSOR_OFF_CHARACTER = 0x17;
  static const uint8_t CLEAR_SCREEN_CHARACTER = 0x0E;
  static const uint8_t HOME_CURSOR_CHARACTER = 0x0C;
  static const uint8_t CURSOR_BACKSPACE_CHARACTER = 0x08; // Moves cursor one space leftward. 
  static const uint8_t CURSOR_TAB_CHARACTER = 0x09; // Moves cursor one space rightward. 
  static const uint8_t CARRIAGE_RETURN_CHARACTER = 0x0D; // Moves cursor to beginning of row.
  static const uint8_t LINE_FEED_CHARACTER = 0x0A; // Moves cursor down one row.
  static const uint8_t CHARACTER_SET_INTERNATIONAL_CHARACTER = 0x18;
  static const uint8_t CHARACTER_SET_JAPANESE_CHARACTER = 0x19;
  static const uint8_t CHARACTER_SET_CODE_PAGE_850_CHARACTER = 0x1A;
  static const uint8_t VERTICAL_SCROLL_OFF_CHARACTER = 0x11;
  static const uint8_t VERTICAL_SCROLL_ON_CHARACTER = 0x12;
  
  static const uint8_t ESC_CHARACTER = 0x1B; // Some commands are prefixed with an escape character.
  static const uint8_t DISPLAY_DIAGNOSTIC_SECOND_CHARACTER = 0x04; // Runs a diagnostic sequence. See data sheet.
  static const uint8_t DISPLAY_ON_SECOND_CHARACTER = 0x05; // Turns display back on after off command has been sent.
  static const uint8_t DISPLAY_OFF_SECOND_CHARACTER = 0x06; // Turns display off. Current display state is retained.

  // When not using the busy pin, the sketch must wait this amount of time
  // for the startup sequence to complete.
  static const uint16_t STARTUP_TIME_IN_MILLISECONDS = 5000; // Measured startup time = 4007 us.
	
  // May be used for busyPin and data7Pin.
  static const uint8_t DO_NOT_USE_PIN = 0xFF;

  // Setting busyPin to something other than DO_NOT_USE_PIN is recommended unless 
  // the application is completely out of unused pins. Otherwise Ncr4X20Vfd will 
  // always assume that the display is ready to accept data, possibly reducing 
  // robustness of the implementation. If busyPin is set to DO_NOT_USE_PIN, the user 
  // sketch must take into account that the display is busy for several seconds 
  // when first powered up. In addition, in performance critical applications the
  // extra delays made by Ncr4X20Vfd to account being unable to wait for 
  // not-busy may be an issue.
  
  // When setting data7Pin to DO_NOT_USE_PIN, Ncr4X20Vfd assumes that the 
  // pin has been tied to ground with a 330 ohm pull down resistor. It will
  // not be possible to send character values 128 through 255 to the display 
  // when data7Pin is set to DO_NOT_USE_PIN. Ncr4X20Vfd will send an alternate
  // character in its place.
  
  // It is possible to share data pins with other devices. Note that Ncr4X20Vfd
  // sets all data pins back to inputs after writing to the display. This may
  // cause issues for other libraries. 

  Ncr4X20Vfd(uint8_t strobePin, uint8_t busyPin,
             uint8_t data0Pin, uint8_t data1Pin, uint8_t data2Pin, uint8_t data3Pin,
             uint8_t data4Pin, uint8_t data5Pin, uint8_t data6Pin, uint8_t data7Pin);
             
  // Provided primarily for consistency with LiquidCrystal API.
  void begin();
  
  void clear();
  void home();
  
  void noDisplay();
  void display();
  void noBlink();
  void blink();
  // Cursor is identical to blink with this VFD. 
  // Could be emulated to match LCD underscore cursor but would
  // add a lot of complexity, memory usage and performance overhead.
  void noCursor();
  void cursor();
  
  // These LiquidCrysal methods are not currently supported.
  // They could be implemented however since they are not
  // native functionality of the NCR display, an implementation
  // would consume substantial code and data memory.
  //
  //void scrollDisplayLeft();
  //void scrollDisplayRight();
  //void leftToRight();
  //void rightToLeft();
  //void autoscroll();
  //void noAutoscroll();  
  
  // It does not appear to be possible to create custom characters with the NCR LCD.
  //void createChar(uint8_t, uint8_t[]);
  void setCursor(uint8_t column, uint8_t row); 
  virtual void write(uint8_t character);
  
  // Additional methods not supported by LiquidCrystal.
  void setDefaultCharacterSet(); // NCR call this "International".
  void setJapaneseCharacterSet();
  void setCodePage850CharacterSet();
  void noVerticalScroll(); // Default. Disables vertical scrolling (cursor wraps to top left of display(.
  void verticalScroll(); // Enables vertical scrolling.
  void startDiagnosticSequence(); // Displays the startup sequence followed by sequencing the entire character set.
  void endDiagnosticSequence(); // Displays the startup sequence followed by sequencing the entire character set.
	
  bool getError();
  void reInitStartupTimeouts(); // Call if sketch powers down/up display via a relay, high side switch etc.
	
private:
  
  enum Flag
  {
    FLAG_BEGIN_HAS_BEEN_CALLED = 0,
    FLAG_HAVE_ESCAPE_COLLISION,
    FLAG_HAD_RESET_TIMEOUT_ERROR,
    FLAG_HAD_WAIT_FOR_NOT_BUSY_TIMEOUT_ERROR,
    
    NUM_FLAGS
  };
  
  static const uint16_t HARD_POWERUP_TIME_IN_MILLISECONDS = 100; // Time display takes to stabalize the lines after a hard powerup.
  static const uint16_t WAIT_FOR_NOT_BUSY_TIMEOUT_IN_MICROSECONDS = 1000;
  static const uint16_t DATA_SETUP_TIME_IN_MICROSECONDS = 1;
  static const uint16_t STROBE_HOLD_TIME_IN_MICROSECONDS = 1;
  static const uint16_t NO_BUSY_PIN_WAIT_TIME_IN_MICROSECONDS = 250; // Measured time was ~128 us.
  
  // Disable default constructors.
  Ncr4X20Vfd();
  Ncr4X20Vfd(const Ncr4X20Vfd&);
  Ncr4X20Vfd& operator=(const Ncr4X20Vfd&);
    
  void enterCommand();
  void leaveCommand();
  void handleSimpleOneCharCommand(uint8_t character);
  void handleSimpleTwoCharCommand(uint8_t char0, uint8_t char1);
  
  bool waitForNotBusy();
  
  // We need to wait on startup for display to be ready.
  static unsigned long _startupTime;
	
  // We do not bit pack flags since the code bloat is likely 
  // worse than a few extra bytes of data.  
  bool _flagBeginHasBeenCalled;
  bool _flagLastWaitForNotBusyFailed;
  bool _flagMightBeInReset;  
  bool _flagHadTimeoutError;
  
  uint8_t _strobePin;
  uint8_t _busyPin;
  uint8_t _dataPins[8];

  // Keep track of last character so we can detect certain conditions.
  uint8_t _lastCharacter;
  
  friend class Ncr4X20UnitTests;
};
  
#endif // Ncr4X20Vfd_h

