/*
  Ncr4x20Vfd Library - USB Serial Adaptor
 
  Uses the Arduino as an adapter to allow
  a PC running LCD Smartie or LCDProc to 
  display on the VFD.
 
  Configure LCD Smartie to use the 
  diplay plugin crystal.dll and startup
  parameters as Arduino COM port,19200,1
  for example:

    COM3,19200,1

  For LCDProc, edit the server .conf file
  (most folks use LCDd.conf). In the 
  [server] section set Driver=CFontz and in
  the [CFontz] set Device=<usb serial device>
  (usually Device=/dev/ttyUSB0), Size=20x4,
  Speed=19200 and NewFirmware=no. Other
  [CFontz] settings can be left at defaults.
 
  TODO: Insert link to wiring tutoral.
 
  Reference page:
    http://arduino.cc/playground/Main/Ncr4X20VfdUsbSerialLcdSmartieAdapter 
 
  Originally created 12 Feb 2011
 
  This file is in the public domain.
*/

#include <Ncr4X20Vfd.h>

// See Ncr4X20VfdTutorialPins.h for other board defines.
#define NCR_4X20_VFD_ARDUINO_PINOUT
#include "Ncr4X20TutorialPins.h"

#include "CustomCharMap.h"

static const uint16_t INITIAL_CUSTOM_CHARACTER = 0x1EE; // A filled diamond.
static const uint16_t DEFAULT_CUSTOM_CHARACTER = 0x1F8; // 100% filled.
static const uint8_t NUM_ACTIVE_CUSTOM_CHARACTERS = 8;

Ncr4X20Vfd vfd(NCR_4X20_VFD_STROBE_PIN,
               NCR_4X20_VFD_BUSY_PIN,
               NCR_4X20_VFD_DATA0_PIN,
               NCR_4X20_VFD_DATA1_PIN,
               NCR_4X20_VFD_DATA2_PIN,
               NCR_4X20_VFD_DATA3_PIN,
               NCR_4X20_VFD_DATA4_PIN,
               NCR_4X20_VFD_DATA5_PIN,
               NCR_4X20_VFD_DATA6_PIN,
               NCR_4X20_VFD_DATA7_PIN);

uint16_t currentCustomCharacters[NUM_ACTIVE_CUSTOM_CHARACTERS] = 
{
  // Inital character is a filled diamond. 
  // We should not usually see these unless
  // there is an error with custom characters.
  INITIAL_CUSTOM_CHARACTER,
  INITIAL_CUSTOM_CHARACTER,
  INITIAL_CUSTOM_CHARACTER,
  INITIAL_CUSTOM_CHARACTER,
  INITIAL_CUSTOM_CHARACTER,
  INITIAL_CUSTOM_CHARACTER,
  INITIAL_CUSTOM_CHARACTER,
  INITIAL_CUSTOM_CHARACTER
};

// Keep track of the current VFD character set being used.
// This is an optimization so we aren't repeatedly 
uint8_t currentCharacterSet = 0;


uint8_t serialGetCharacter() {

  while (!Serial.available()) {
  }

  Serial.read();
}

// See CustomCharMap.h for more detais on what this function
// does.
uint16_t getBestFitForCustomCharacter(uint8_t *pixels) {

  const uint8_t *currentCharMapPixels = customCharMapPixels;

  for (uint8_t characterIndex = 0; characterIndex < CUSTOM_CHAR_MAP_LEN; characterIndex++)
  {
    bool match = true;

    for (uint8_t i=0; i<7; i++)
    {
      if (*currentCharMapPixels != pixels[i])
      {
        match = false;
      }
      currentCharMapPixels++;
    }

    if (match)
    {
      return customCharMapChars[characterIndex];
    }
  }

  // Return all-filled.
  return DEFAULT_CUSTOM_CHARACTER;    
}

void handleCustomCharacter()
{
  uint8_t characterIndex = serialGetCharacter();
  uint8_t pixels[8];

  for (uint8_t i=0; i<8; i++)
  {
    // Character pixel rows.
    pixels[i] = serialGetCharacter() & 0b11111;
  }

  if (characterIndex < NUM_ACTIVE_CUSTOM_CHARACTERS)
  {
    currentCustomCharacters[characterIndex] =
      getBestFitForCustomCharacter(pixels);
  }
}

void handleCrystalControlCode(uint8_t character) {

  switch (character)
  {  
  case 3:
    // Restore display.
    // Not sure what this is but we'll enable our display.
    vfd.display();
    break;

  case 4:
    // Turn off cursor.
    vfd.noCursor();
    break;

  case 12:
    // Clear screen.
    vfd.clear();
    break;

  case 14: // Set brightness.
  case 15: // Set contrast, range 0-100.
    serialGetCharacter(); // Not implemented: read and discard byte.
    break;

  case 17:
    {
      // Set position.
      vfd.setCursor(serialGetCharacter(), serialGetCharacter());
      break;
    }
  
  case 19:
    // Turn on scroll.
    vfd.verticalScroll();
    break;
  
  case 20:
    // Turn off scroll.
    vfd.noVerticalScroll();
    break;

  case 22:
    // Disable scrolling marquee.
    // Not sure how many bytes, but LCD Smartie is sending 3 more
    // so read and discard all
    serialGetCharacter();
    serialGetCharacter();
    serialGetCharacter();
    break;      

  case 23: // Wrap on.
  case 24: // Wrap off.
    // Ignore, display always wraps.
    break;

  case 25:
    // Set custom character pixels.
    handleCustomCharacter();
    break;

  case 26:
    // Init display.
    vfd.begin();
    break;

    // Pass standard ASCII control codes straight through.
  case 8:
  case 9:
  case 10:
  case 13:
    vfd.write(character);
    break;

  default:
    // Unknown code, ignore it.
    break;    
  }
}

void setup() {

  // Drive the unneeded control lines high. The display's parallel 
  // interface will not function correctly if they are low.
  digitalWrite(NCR_4X20_VFD_RESET_PIN, HIGH);
  digitalWrite(NCR_4X20_VFD_UNKNOWN_CONTROL1_PIN, HIGH);
  digitalWrite(NCR_4X20_VFD_UNKNOWN_CONTROL2_PIN, HIGH);

  pinMode(NCR_4X20_VFD_RESET_PIN, OUTPUT);
  pinMode(NCR_4X20_VFD_UNKNOWN_CONTROL1_PIN, OUTPUT);
  pinMode(NCR_4X20_VFD_UNKNOWN_CONTROL2_PIN, OUTPUT); 

  // Start the library.
  vfd.begin();

  // Set current character set to match the default of the VFD library.
  currentCharacterSet = 0;

  // Initialize the serial connection.
  Serial.begin(19200);
}

void loop() {

  uint8_t character = serialGetCharacter();

  if (character < 32)
  {
    handleCrystalControlCode(character);
  }
  else
  {
    // Translate character to character set/character pair.
    uint16_t translated = character;

    if (character>=128 && character<(128+8))
    {
      translated = currentCustomCharacters[character - 128];
    }

    // Remap selected extended characters.
    switch (character)
    {
      case 0xDE:
        // Up arrow.
        translated = 0x1FC;
        break;
      
      case 0xE0:
        // Down arrow.
        translated = 0x1FA;
        break;
        
      case 0xE1:
        // Left arrow.
        translated = 0x1FD;
        break;
        
      case 0xDF:
        // Right arrow.
        translated = 0x1FB;
        break;
        
      case 0xFF:
        // Filled block.
        translated = 0x1F8;
        break;
        
      default:
        // Leave unchanged.
        break;   
    }

    // Switch character set if has changed.
    uint8_t characterSet = translated >> 8;
    if (characterSet != currentCharacterSet && characterSet < 3)
    {
      // Send character set command directly.
      vfd.write(Ncr4X20Vfd::CHARACTER_SET_INTERNATIONAL_CHARACTER + characterSet);

      currentCharacterSet = characterSet;
    }

    // Write the character now that the character set has been selected.
    vfd.write((uint8_t)translated);
  }  
}


